'''
Utils for Chuda internals
'''

import collections

default_logger_config = {
    "version": 1,
    "formatters": {
        "default": {
            "format": '%(message)s'
        },
        "date": {
            "format": '%(asctime)s - %(levelname)s - %(message)s',
            "datefmt": '%Y-%m-%d %H:%M:%S'
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default",
            "level": "INFO"
        }
    },
    "loggers": {
        "default": {
            "handlers": ["console"],
            "level": "INFO"
        }
    }
}


def dict_merge(dct, merge_dct):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k, _ in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], collections.Mapping)):
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]


def get_flag(base, attr):
    return getattr(base, attr, False)


def isset(var, scope=globals):
    return var in scope()


class Null(object):

    _instances = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Null, cls).__new__(cls, *args, **kwargs)
        return cls._instances[cls]

    def __init__(self, *args, **kwargs):
        pass

    def __repr__(self):
        return "<Null>"

    def __str__(self):
        return ""

    def __eq__(self, other):
        return id(self) == id(other) or other is None

    __nonzero__ = __bool__ = lambda self: False

    nullify = lambda self, *x, **kwargs: self

    __call__ = nullify
    __getattr__ = __setattr__ = __delattr__ = nullify
    __cmp__ = __ne__ = __lt__ = __gt__ = __le__ = __ge__ = nullify
    __pos__ = __neg__ = __abs__ = __invert__ = nullify
    __add__ = __sub__ = __mul__ = __mod__ = __pow__ = nullify
    __floordiv__ = __div__ = __truediv__ = __divmod__ = nullify
    __lshift__ = __rshift__ = __and__ = __or__ = __xor__ = nullify
    __radd__ = __rsub__ = __rmul__ = __rmod__ = __rpow__ = nullify
    __rfloordiv__ = __rdiv__ = __rtruediv__ = __rdivmod__ = nullify
    __rlshift__ = __rrshift__ = __rand__ = __ror__ = __rxor__ = nullify
    __iadd__ = __isub__ = __imul__ = __imod__ = __ipow__ = nullify
    __ifloordiv__ = __idiv__ = __itruediv__ = __idivmod__ = nullify
    __ilshift__ = __irshift__ = __iand__ = __ior__ = __ixor__ = nullify
    __getitem__ = __setitem__ = __delitem__ = nullify
    __getslice__ = __setslice__ = __delslice__ = nullify
    __reversed__ = nullify
    __contains__ = __missing__ = nullify
    __enter__ = __exit__ = nullify



'''
Module for Chuda Plugins
'''
class Plugin:
    '''
    Class represent a Plugin for a Chuda application
    A plugin can register hooks for import step in application lifecycle
    or enrich the app with new properties
    '''
    priority = 0

    def __init__(self):
        self.app = None

    def setup(self, app):
        '''
        Setup the app on the plugin
        '''
        self.app = app

    def enrich_app(self, name, value):
        '''
        Add a new property to the app
        '''
        #Method shouldn't be added:  https://stackoverflow.com/a/28060251/3042398
        if type(value) == type(self.enrich_app):
            raise ValueError("enrich_app can't add method")

        setattr(self.app, name, value)

    def on_create(self):
        '''
        Called at the app creation
        '''
        pass

    def on_arguments_parsed(self):
        '''
        Called after arguments have been parsed
        '''
        pass

    def on_config_read(self):
        '''
        Called after configurations files have been read
        '''
        pass

    def on_signals_handled(self):
        '''
        Called after signals handlers have been setuped
        '''
        pass

    def on_logger_created(self):
        '''
        Called after logger have been created
        '''
        pass

    def on_run(self):
        '''
        Called just before run the app
        '''
        pass

    def on_end(self):
        '''
        Called after all steps and code have been executed
        '''
        pass



'''
'''

import sys
from .app import App

class Command(App):
    '''
    A subcommand for multicommands cli tool
    '''
    command_name = None
    parent_config = None
    use_subconfig = False
    merge_parent_arguments = True

    def __str__(self):
        return "<ChudaCommand command_name={}>".format(self.command_name)

    def __repr__(self):
        return "<ChudaCommand command_name={}>".format(self.command_name)

    def __init__(self): # pylint: disable=W0231
        pass

    def setup(self, parent_arguments, parent_config, parent_logger):
        '''
        Setup properties from parent app on the command
        '''
        self.logger = parent_logger

        if self.command_name is None:
            self.logger.error("The command_name attribute is required")
            sys.exit(1)

        self.parent_config = parent_config
        self.arguments = parent_arguments

        if self.use_subconfig:
            self.__init_config()
        else:
            self.config = self.parent_config


'''
Chuda is a simple framework to create CLI tools
'''

from .app import App
from .commands import Command
from .decorators import autorun, signal_handler
from .plugins import Plugin
from .arguments import Option, Parameter



import os
import argparse
import logging
import logging.config
import importlib
import signal
import delegator
import argcomplete
import chuda.utils as utils
from .arguments import Option
from .plugins import Plugin


class App:
    app_name = None
    arguments = []
    config = {}
    config_parser = "ini"
    config_path = None
    default_arguments = [
        Option(name=["-q", "--quiet"], dest="quiet", action="store_true"),
        Option(name=["-v", "--verbose"], dest="verbose", action="store_true"),
        Option(name=["--version"], dest="version", action="store_true")
    ]
    description = None
    logger = None
    override_default_arguments = False
    parser = None
    plugins = []
    shell = delegator
    subcommands = []
    version = "0.0.1"

    __signal_handlers = {}

    def __str__(self):
        return "<ChudaApp app_name={}>".format(self.app_name)

    def __repr__(self):
        return "<ChudaApp app_name={}>".format(self.app_name)

    def in_autocomplete_mode(self):
        return "_ARGCOMPLETE" in os.environ

    def __init_arguments(self):
        self.parser = argparse.ArgumentParser(
            prog=self.app_name,
            description=self.description
        )

        self.parser.set_defaults(command="main")

        if not self.override_default_arguments:
            self.arguments = self.default_arguments + self.arguments

        for argument in self.arguments:
            arg_tuple = argument.convert_to_argument()
            if isinstance(arg_tuple[0], list):
                parg = self.parser.add_argument(*arg_tuple[0], **arg_tuple[1])
            else:
                parg = self.parser.add_argument(arg_tuple[0], **arg_tuple[1])

            if argument.completer:
                parg.completer = argument.completer


        subcommands_dict = {}
        if self.subcommands:
            subparsers = self.parser.add_subparsers(
                title="subcommands"
            )
            
        for subcommand in self.subcommands:
            instance = subcommand()
            subcommands_dict[instance.command_name] = instance
            subparser = subparsers.add_parser(
                instance.command_name,
                help=instance.description,
                description=instance.description
            )
            subparser.set_defaults(command=instance.command_name)
            if instance.merge_parent_arguments:
                instance.arguments = self.arguments + instance.arguments

            for argument in instance.arguments:
                arg_tuple = argument.convert_to_argument()
                if isinstance(arg_tuple[0], list):
                    sp_arg = subparser.add_argument(*arg_tuple[0], **arg_tuple[1])
                else:
                    sp_arg = subparser.add_argument(arg_tuple[0], **arg_tuple[1])

                if argument.completer:
                    sp_arg.completer = argument.completer

        argcomplete.autocomplete(
            argument_parser=self.parser,
            always_complete_options=False
        )
        self.arguments = self.parser.parse_args()
        self.subcommands = subcommands_dict

    def __init_config(self):
        if self.config_path:
            config_path_list = []
            if isinstance(self.config_path, str):
                config_path_list = [self.config_path]
            elif isinstance(self.config_path, list):
                config_path_list = self.config_path

            for path in config_path_list:
                if os.path.isfile(path):
                    self.config_path = path
                    break
                else:
                    self.config_path = None

            if self.config_path:
                if self.config_parser.lower() == "ini":
                    configparser = importlib.import_module("configparser")
                    tmp = configparser.ConfigParser()
                    tmp.read(self.config_path)
                    for section in tmp.sections():
                        utils.dict_merge(self.config, {section: tmp[section]})

                if self.config_parser.lower() == "yaml":
                    yaml = importlib.import_module("yaml")
                    with open(self.config_path, "r") as file:
                        tmp = yaml.load(file.read())
                        utils.dict_merge(self.config, tmp)

                if self.config_parser.lower() == "json":
                    json = importlib.import_module("json")
                    with open(self.config_path, "r") as file:
                        tmp = json.load(file)
                        utils.dict_merge(self.config, tmp)

    def __init_logging(self):
        logging_config = utils.default_logger_config
        if self.config.get("logging", None):
            utils.dict_merge(logging_config, self.config["logging"])

        if utils.get_flag(self.arguments, "verbose"):
            for name, handler in logging_config.get("handlers", {}).items():
                handler["level"] = "DEBUG"
                logging_config["handlers"][name] = handler

            for name, logger in logging_config.get("loggers", {}).items():
                logger["level"] = "DEBUG"
                logging_config["loggers"][name] = logger

        self.config["logging"] = logging_config
        logging.config.dictConfig(self.config["logging"])

        if utils.get_flag(self.arguments, "quiet"):
            self.logger = utils.Null()
        else:
            self.logger = logging.getLogger("default")

    def __init_plugins(self):
        instances = []

        for plugin in self.plugins:
            if isinstance(plugin, type):
                instance = plugin()
                if not isinstance(plugin, Plugin):
                    raise TypeError("plugins should subclasse the Plugin class")
                instance.setup(self)
                instances.append(instance)
            elif isinstance(plugin, Plugin):
                plugin.setup(self)
                instances.append(plugin)
            else:
                raise TypeError("plugins should subclasse the Plugin class")

        instances.sort(key=lambda p: p.priority)

        self.plugins = instances

    def call_plugins(self, step):
        '''
        Check if a 'step' method exist on each plugin, and call it
        '''
        for plugin in self.plugins:
            try:
                getattr(plugin, step)()
            except AttributeError:
                self.logger.debug("{} doesn't exist on plugin {}".format(step, plugin))
            except TypeError:
                self.logger.debug("{} on plugin {} is not callable".format(step, plugin))

    def __init_signals(self):
        for attribute_name in dir(self):
            attribute = getattr(self, attribute_name)
            handle_signal = getattr(attribute, "handle_signal", None)
            if handle_signal:
                self.__signal_handlers \
                         .setdefault(handle_signal, []) \
                         .append(attribute)

        for plugin in self.plugins:
            for attribute_name in dir(plugin):
                attribute = getattr(plugin, attribute_name)
                handle_signal = getattr(attribute, "handle_signal", None)
                if handle_signal:
                    self.__signal_handlers \
                        .setdefault(handle_signal, []) \
                        .append(attribute)

        for key in self.__signal_handlers:
            signal.signal(key, self.__signal_handler_factory(key))

    def __signal_handler_factory(self, sig):
        def __signal_handler(signum, frame):
            for handler in self.__signal_handlers[sig]:
                handler(signum, frame)

        return __signal_handler

    def __init__(self):
        self.__init_plugins()
        self.call_plugins("on_create")
        self.__init_signals()
        self.call_plugins("on_signals_handled")
        self.__init_config()
        self.call_plugins("on_config_read")
        self.__init_arguments()
        self.call_plugins("on_arguments_parsed")
        self.__init_logging()
        self.call_plugins("on_logger_created")

    def run(self):
        self.call_plugins("on_run")
        if vars(self.arguments).get("version", None):
            self.logger.info("{app_name}: {version}".format(app_name=self.app_name, version=self.version))
        else:
            if self.arguments.command == "main":
                self.main()
            else:
                command = self.subcommands[self.arguments.command]
                command.setup(self.arguments, self.config, self.logger)
                command.main()
        self.call_plugins("on_end")

    def main(self):
        if not utils.get_flag(self.arguments, "quiet"):
            self.parser.print_help()



'''
Module with argparse utils for chuda
'''


class Argument:
    '''
    Abstract parent class for Option and Parameter
    '''
    name = None
    action = "store"
    nargs = None
    const = None
    default = None
    type = None
    choices = None
    required = None
    help = None
    metavar = None
    dest = None
    completer = None

    def __init__(self, name=None, action="store", nargs=None, const=None,
                 default=None, type=None, choices=None, required=None, help=None,
                 metavar=None, dest=None, completer=None):
        self.name = name
        self.action = action
        self.nargs = nargs
        self.const = const
        self.default = default
        self.type = type
        self.choices = choices
        self.required = required
        self.help = help
        self.metavar = metavar
        self.dest = dest
        self.completer = completer

    def convert_to_argument(self):
        field_list = [
            "action", "nargs", "const", "default", "type",
            "choices", "required", "help", "metavar", "dest"
        ]

        return (
            self.name,
            {
                field: getattr(self, field) for field in field_list if getattr(self, field) is not None
            }
        )


class Option(Argument):
    '''
    Represent an option on the command-line (--whatever)
    '''

    def __repr__(self):
        return "<Option name={}>".format(self.name)

    def __str__(self):
        return "<Option name={}>".format(self.name)

    def convert_to_argument(self):
        if not isinstance(self.name, list) and not isinstance(self.name, tuple):
            raise TypeError("Inapropriate name type for Option. Should be a tuple or a list")

        return super(Option, self).convert_to_argument()


class Parameter(Argument):
    '''
    Represent a parameter on the command-line (macommand toto)
    '''

    def __repr__(self):
        return "<Parameter name={}>".format(self.name)

    def __str__(self):
        return "<Parameter  name={}>".format(self.name)

    def convert_to_argument(self):
        if not isinstance(self.name, str):
            raise TypeError("Inapropriate name type for Parameter. Should be a string")

        return super(Parameter, self).convert_to_argument()



from setuptools import setup

try:
    with open('README.MD') as readme_file:
        README = readme_file.read()
except FileNotFoundError:
    README = ""

VERSION = '0.0.12'

setup(
    name='chuda',
    packages=['chuda'],
    version=VERSION,
    description='A simple framework to create CLI tools',
    long_description=README,
    license='MIT',
    install_requires=[
        "delegator.py>=0.1.0",
        "argcomplete>=1.9.4",
    ],
    author='Romain Moreau',
    author_email='moreau.romain83@gmail.com',
    url='https://github.com/Varkal/chuda',
    download_url='https://github.com/Varkal/chuda/archive/{}.tar.gz'.format(VERSION),
    keywords=['cli', 'chuda'],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Programming Language :: Python :: 3"
    ],
)



'''
Module contains Chuda decorators
'''
def autorun():
    '''
    This decorator automaticaly call the run method of this class if
    it's in the main file
    '''
    def wrapper(cls):
        '''
        Wrapper for autorun
        '''

        import inspect
        if inspect.getmodule(cls).__name__ == "__main__":
            cls().run()
        return cls

    return wrapper


def signal_handler(sig):
    '''
    Flag a method to be used as a signal handler
    '''
    def wrapper(func):
        '''
        Wrapper for signal_handler
        '''
        setattr(func, "handle_signal", sig)
        return func

    return wrapper
